#!/usr/bin/env python3

import os
import bz2
import numpy as np
import matplotlib.pyplot as plt
import datetime
from ftplib import FTP
from scipy.optimize import leastsq
from os.path import join
plt.rcParams['keymap.pan'] = ''
plt.rcParams['toolbar'] = 'None'

strTime = '1670407303' #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
resDir = 'E:/result/' #<<<<<<<<<<A local folder
plotSiteNum = 13
delPPS = 1 #1: delete PPS signal; 0: do not delete
loc2D = 1 #1: 2d locating; 0: 3d locating


strNatTime = datetime.datetime.fromtimestamp(float(strTime)).strftime('%Y%m%d-%H%M%S')
strDate = strNatTime[0:8]
print(strNatTime)

if strDate>'20181101' and strDate<'20190401':
    dataRoot = '/media/lightning/HDD1/2018-2/LF/'+strDate+'/'
    oriDataLen = 1000000
    lengthFactor = 1
    year = 2018
elif strDate>'20191101' and strDate<'20200401':
    dataRoot = '/media/lightning/HDD1/2019/LF/'+strDate+'/'
    oriDataLen = 5000000
    lengthFactor = 5
    year = 2019
elif strDate>'20201001' and strDate<'20210601':
    dataRoot = '/media/lightning/HDD2/2020/LF/'+strDate+'/'
    oriDataLen = 1000000
    lengthFactor = 1
    year = 2020
elif strDate>='20210601' and strDate<'20220401':
    dataRoot = '/media/lightning/HDD3/2021/'+strDate+'/'
    oriDataLen = 1000000
    lengthFactor = 1
    year = 2021
elif strDate>='20220401' and strDate<'20220901':
    dataRoot = '/media/lightning/HDD4/2022/'+strDate+'/'
    oriDataLen = 1000000
    lengthFactor = 1
    year = 2021
elif strDate>='202200901':
    dataRoot = '/media/lightning/HDD5/2022/'+strDate+'/'
    oriDataLen = 1000000
    lengthFactor = 1
    year = 2021
else:
    print('Error: No data available!')
    
ftp_add = '10.226.82.10'
ftp_port = 21
ftp_account = 'lightning'
ftp_key = 'discharge'
    
peakWidth = 10*lengthFactor
upWidth = 20*lengthFactor
downWidth = 20*lengthFactor
slideLen = 10*lengthFactor

cAllSite = [       'AKK',     'FTK',     'GFU',    'HIM',    'HKI',      'IBG',    'IMZ',      'KGM',      'KSM',    'KZK',      'MIN',      'NAT',      'NNO',     'SSK',     'SUM',      'TBT',     'TKO',     'TYM']
cAllY    = [    36.42482,  37.14248, 35.464567, 36.85695, 36.89355,   35.63635, 36.73055,    36.5178,  35.715483, 36.60805,  36.666517,  36.552833,   37.04315,  36.44767, 36.665283,  36.669217, 36.707583, 36.752083]
cAllX    = [   136.42265, 136.69226,136.738367,  136.973, 136.7791, 136.484567, 137.0756, 136.742883,   137.3707,  136.594, 136.665233, 136.954983, 136.966983, 136.64967, 136.64515, 136.728717,  136.9237,137.297100]
cAllZ  =       [      13,        36,        44,       17,       46,        230,       45,        423,        555,       15,          6,        120,         31,       659,        15,         23,        42,        12]

allSiteNum = len(cAllSite)
mainSiteMat = ['HKI', 'SSK', 'AKK', 'FTK', 'MIN']

cAllX = np.array(cAllX)/180.0*np.pi
cAllY = np.array(cAllY)/180.0*np.pi
cAllZ = np.array(cAllZ)

global dataMatList, xMat, drawSiteMat
global showDataMatList, showXMat
global curIndex
global linePola, lineMat

linePola = 1
lineMat = [-1 for i in range(plotSiteNum)]
    
#Make figure
f = plt.figure(figsize=(16,12))

belowMargin = 0.06
aboveMargin = 0.02
height = (1-belowMargin-aboveMargin)/plotSiteNum
width = 0.9
leftMargin = 0.06
rightMargin = 1-width-leftMargin

axMat = list(range(plotSiteNum))
for i in range(plotSiteNum):
    axMat[i] = f.add_axes([leftMargin, belowMargin+height*(plotSiteNum-1-i), width, height])
    if i<plotSiteNum-1:
        plt.setp(axMat[i].get_xticklabels(), visible=False)

def getDivideLen(totalLen):
    if totalLen>1e7:
        divideLen = 100
    elif totalLen>1e6:
        divideLen = 10
    else:
        divideLen = 1
    return divideLen

def delPPSFun(dataMat):
    dataMat = dataMat.copy()
    zeroVal = np.median(dataMat)
    
    curIndex = 0
    while True:
        if dataMat[curIndex]<1000+zeroVal:
            break
        curIndex += 1
    preVal = dataMat[curIndex]
    while True:
        curIndex += 1
        curVal = dataMat[curIndex]
        if curVal>preVal:
            break
        preVal = curVal
    dataMat[0:curIndex+lengthFactor] = dataMat[curIndex+lengthFactor]
    
    ppsMinIndex = dataMat[200000*lengthFactor:2000020*lengthFactor].argmin()+200000*lengthFactor
    curIndex = ppsMinIndex
    while True:
        if dataMat[curIndex]>-1000+zeroVal:
            break
        curIndex -= 1
    preVal = dataMat[curIndex]
    while True:
        curIndex -= 1
        curVal = dataMat[curIndex]
        if curVal<preVal:
            break
        preVal = curVal
    dataMat[curIndex:ppsMinIndex] = dataMat[curIndex-1]
    
    curIndex = ppsMinIndex
    while True:
        if dataMat[curIndex]>-1000+zeroVal:
            break
        curIndex += 1
    preVal = dataMat[curIndex]
    while True:
        curIndex += 1
        curVal = dataMat[curIndex]
        if curVal<preVal:
            break
        preVal = curVal
    dataMat[ppsMinIndex:curIndex+2*lengthFactor] = dataMat[curIndex+2*lengthFactor]
    
    curIndex = 1000000*lengthFactor-1
    preVal = dataMat[curIndex]
    while True:
        curIndex -= 1
        curVal = dataMat[curIndex]
        if curVal>preVal:
            break
        preVal = curVal
    dataMat[curIndex:1000000*lengthFactor] = dataMat[curIndex-1]

    return dataMat

def readData():
    global dataMatList, xMat, drawSiteMat, dataSiteNum
    
    #Download data
    ftp = FTP()
    ftp.connect(ftp_add, ftp_port)
    ftp.login(ftp_account, ftp_key)
    ftp.cwd(dataRoot)
    siteMat = ftp.nlst()
    for siteName in cAllSite:
        if siteName not in siteMat:
            continue
        ftpDir = dataRoot+siteName+'/'
        fileName = siteName+'L'+strTime+'.bz2'
        ftp.cwd(ftpDir)
        fidFile = open(join(resDir, fileName), 'wb')
        try:
            ftp.retrbinary('RETR '+fileName, fidFile.write)
        except:
            fidFile.close()
            os.remove(join(resDir, fileName))
            continue
        fidFile.close()
    ftp.close()
    
    #Read data
    dataMatList = []
    drawSiteMat = []
    for siteName in cAllSite:
        fileName = siteName+'L'+strTime+'.bz2'
        fullFileName = join(resDir, fileName)
        if os.path.exists(fullFileName)==False:
            continue
        try:
            fidFile = bz2.BZ2File(fullFileName)
            oriDataMat = np.frombuffer(fidFile.read(), np.int16)
            if len(oriDataMat)!=oriDataLen:
                fidFile.close()
                os.remove(join(resDir, fileName))
                continue
            fidFile.close()
            os.remove(join(resDir, fileName))
        except Exception as e:
            print(str(e))
            fidFile.close()
            os.remove(join(resDir, fileName))
            continue
        
        if (year==2019 or year==2020) and delPPS==1:
            oriDataMat = delPPSFun(oriDataMat)
        drawSiteMat.append(siteName)
        dataMatList.append(oriDataMat)
    
    xMat = np.arange(oriDataLen)
    dataSiteNum = len(drawSiteMat)
    if dataSiteNum>plotSiteNum:
        dataSiteNum = plotSiteNum
    print(drawSiteMat)

def drawPro(centerX, xLen):#xLen is the points at one side
    global showDataMatList, showXMat
    showDataMatList = []
    
    for i in range(plotSiteNum):
        axMat[i].clear()
    
    divideLen = getDivideLen(xLen*2)
    startX = int(centerX-xLen)
    if startX<0:
        startX = 0
    endX = int(centerX+xLen)
    if endX>oriDataLen:
        endX = oriDataLen
    showXMat = xMat[startX:endX]
    
    for i in range(dataSiteNum):
        siteName = drawSiteMat[i]
        showDataMat = dataMatList[i][startX:endX]
        showDataMatList.append(showDataMat)
        axMat[i].plot(showXMat[0::divideLen], showDataMat[0::divideLen])
        axMat[i].text(0.98, 0.98, siteName, horizontalalignment='right', verticalalignment='top', transform = axMat[i].transAxes)
    axMat[dataSiteNum//2].set_ylabel('E (Digital Unit)')
    axMat[plotSiteNum-1].set_xlabel('Data points')
    for i in range(plotSiteNum):
        axMat[i].set_xlim([startX, endX])
    f.canvas.draw()

def plotLine():
    global lineMat
    for i in range(len(drawSiteMat)):
        showDataMat = showDataMatList[i]
        if linePola>0:
            lineX = showDataMat.argmax()
        else:
            lineX = showDataMat.argmin()
        yMin, yMax = axMat[i].get_ylim()
        peakLine, = axMat[i].plot([lineX+showXMat[0], lineX+showXMat[0]], [yMin, yMax], 'g')
        lineMat[i] = peakLine
    f.canvas.draw()

def chanPola():
    global linePola
    linePola = linePola*-1
    for i in range(len(drawSiteMat)):
        showDataMat = showDataMatList[i]
        if linePola>0:
            lineX = showDataMat.argmax()
        else:
            lineX = showDataMat.argmin()
        lineMat[i].set_xdata([lineX+showXMat[0], lineX+showXMat[0]])
    f.canvas.draw()

def delLine(ax):
    axIndex = axMat.index(ax)
    lineMat[axIndex].remove()
    lineMat[axIndex] = -1
    f.canvas.draw()

def yIncre(ax, curY):
    yMin, yMax = ax.get_ylim()
    ylimLen = yMax-yMin
    newMin = curY-ylimLen*0.45
    newMax = curY+ylimLen*0.45
    ax.set_ylim([newMin, newMax])
    f.canvas.draw()
    
def yIncreAll():
    for ax in axMat:
        yMin, yMax = ax.get_ylim()
        newMin = yMin*0.6
        newMax = yMax*0.6
        ax.set_ylim([newMin, newMax])
    f.canvas.draw()
    


def calPeak(latiVal, longiVal):
    convertCoefMat = [    1, 0.294, 0.876, 1.323, 2.362, 0.867, 1.273, 0.364]
    convertSiteMat = ['MIN', 'TKO', 'IMZ', 'HKI', 'KGM', 'NAT', 'NNO', 'TBT']
    convertCoefMat2 = [0.748, 1.135, 0.694, 0.780, 0.532]
    convertSiteMat2 = ['KZK', 'SSK', 'FTK', 'HIM', 'AKK']
    
    siteMat = []
    peakMat = []
    disMat = []
    for i in range(dataSiteNum):
        if type(lineMat[i]) is int:
            continue
        peakX = lineMat[i].get_xdata()[0]
        siteName = drawSiteMat[i]
        siteIndex = cAllSite.index(siteName)
        siteLongi = cAllX[siteIndex]
        siteLati = cAllY[siteIndex]
        siteDis = dis(latiVal, longiVal, siteLati, siteLongi)/1000
        if siteDis<40:
            continue
        dataMat = dataMatList[i]
        peakVal = dataMat[peakX]
        siteMat.append(siteName)
        peakMat.append(peakVal)
        disMat.append(siteDis)
    
    norAmpMat = []
    selSiteMat = []
    for i in range(len(convertSiteMat)):
        siteName = convertSiteMat[i]
        if siteName in siteMat:
            selSiteMat.append(siteName)
            siteIndex = siteMat.index(siteName)
            peakVal = peakMat[siteIndex]
            norPeakVal = peakVal*disMat[siteIndex]/100/convertCoefMat[i]
            norAmpMat.append(norPeakVal)
            # print(siteName, norPeakVal)
        if len(norAmpMat)>=3:
            break
    
    if len(norAmpMat)==0:
        for i in range(len(convertSiteMat2)):
            siteName = convertSiteMat2[i]
            if siteName in siteMat:
                selSiteMat.append(siteName)
                siteIndex = siteMat.index(siteName)
                peakVal = peakMat[siteIndex]
                norPeakVal = peakVal*disMat[siteIndex]/100/convertCoefMat2[i]
                norAmpMat.append(norPeakVal)
                # print(siteName, norPeakVal)
    
    if len(norAmpMat)>0:
        return -int(round(sum(norAmpMat)/len(norAmpMat)*0.00549))
    else:
        return 0


def calPeak2021(latiVal, longiVal):
    convertCoefMat = [    1, 0.809, 1.857]
    convertSiteMat = ['GFU', 'KSM', 'MIN']
    convertCoefMat2 = [1.628, 4.355, 0.732, 4.033, 1.224, 1.133, 0.210, 1.267]
    convertSiteMat2 = ['AKK', 'FTK', 'HIM', 'HKI', 'NAT', 'NNO', 'SUM', 'TYM']
    
    siteMat = []
    peakMat = []
    disMat = []
    for i in range(dataSiteNum):
        if type(lineMat[i]) is int:
            continue
        peakX = lineMat[i].get_xdata()[0]
        siteName = drawSiteMat[i]
        siteIndex = cAllSite.index(siteName)
        siteLongi = cAllX[siteIndex]
        siteLati = cAllY[siteIndex]
        siteDis = dis(latiVal, longiVal, siteLati, siteLongi)/1000
        if siteDis<40:
            continue
        dataMat = dataMatList[i]
        peakVal = dataMat[peakX]
        siteMat.append(siteName)
        peakMat.append(peakVal)
        disMat.append(siteDis)
    
    norAmpMat = []
    selSiteMat = []
    for i in range(len(convertSiteMat)):
        siteName = convertSiteMat[i]
        if siteName in siteMat:
            selSiteMat.append(siteName)
            siteIndex = siteMat.index(siteName)
            peakVal = peakMat[siteIndex]
            norPeakVal = peakVal*disMat[siteIndex]/100/convertCoefMat[i]
            norAmpMat.append(norPeakVal)
        if len(norAmpMat)>=3:
            break
    
    if len(norAmpMat)==0:
        for i in range(len(convertSiteMat2)):
            siteName = convertSiteMat2[i]
            if siteName in siteMat:
                selSiteMat.append(siteName)
                siteIndex = siteMat.index(siteName)
                peakVal = peakMat[siteIndex]
                norPeakVal = peakVal*disMat[siteIndex]/100/convertCoefMat2[i]
                norAmpMat.append(norPeakVal)
    
    if len(norAmpMat)>0:
        return -int(round(sum(norAmpMat)/len(norAmpMat)*0.01087))
    else:
        return 0
    

R = 6370519
#>>>>>>>>>Target function for optimization
speed = 300/lengthFactor # m/point
def minFcn(X0, xi, yi, zi, ti):
    h0 = R+X0[3]
    h1 = R+zi
    angleCoef = np.sin(X0[1])*np.sin(yi)+np.cos(X0[1])*np.cos(yi)*np.cos(X0[0]-xi)
    l = (h0**2+h1**2-2*h0*h1*angleCoef)**0.5
    return l-(speed*(ti-X0[2])) #Should always be ti-X0[2] (Not X0[2]-ti) to make sure the result of X0[2] will always smaller than ti.
#<<<<<<<<<<<<<

def minFcn2D(X0, xi, yi, zi, ti):
    h0 = R+1000
    h1 = R+zi
    angleCoef = np.sin(X0[1])*np.sin(yi)+np.cos(X0[1])*np.cos(yi)*np.cos(X0[0]-xi)
    l = (h0**2+h1**2-2*h0*h1*angleCoef)**0.5
    return l-(speed*(ti-X0[2])) #Should always be ti-X0[2] (Not X0[2]-ti) to make sure the result of X0[2] will always smaller than ti.
#<<<<<<<<<<<<<

def locFun(xi, yi, zi, ti): 
    dataSiteNum = len(xi)
    #Initial value
    minIndex = ti.argmin()
    iniPosiX = xi[minIndex]
    iniPosiY = yi[minIndex]
    iniPosiZ = 2000
    #Optimization
    X_ini = np.array([iniPosiX, iniPosiY, ti.min()-100, iniPosiZ])
    resMat, success = leastsq(minFcn, X_ini, args=(xi, yi, zi, ti))
    #Residual
    aveResi = 0
    resiMat = []
    for j in range(dataSiteNum):
        resiVal = abs(minFcn(resMat, xi[j], yi[j], zi[j], ti[j]))
        resiMat.append(resiVal)
        aveResi += resiVal**2
    aveResi = (aveResi/dataSiteNum)**0.5 #m
    return [resMat, resiMat, aveResi]

def locFun2D(xi, yi, zi, ti): 
    dataSiteNum = len(xi)
    #Initial value
    minIndex = ti.argmin()
    iniPosiX = xi[minIndex]
    iniPosiY = yi[minIndex]
    #Optimization
    X_ini = np.array([iniPosiX, iniPosiY, ti.min()-100])
    resMat, success = leastsq(minFcn2D, X_ini, args=(xi, yi, zi, ti))
    #Residual
    aveResi = 0
    resiMat = []
    for j in range(dataSiteNum):
        resiVal = abs(minFcn2D(resMat, xi[j], yi[j], zi[j], ti[j]))
        resiMat.append(resiVal)
        aveResi += resiVal**2
    aveResi = (aveResi/dataSiteNum)**0.5 #m
    return [resMat, resiMat, aveResi]

lati0 = 36.76/180*np.pi
longi0 = 136.76/180*np.pi
def convertXY(longiVal, latiVal):
    #x in longitude direction, y in latitude direction
    y = (latiVal-lati0)*R
    x = 2*R*np.arcsin(np.sin((longiVal-longi0)/2.0)*np.cos(lati0))
    return [x, y]

def calLoc():
    xMat = np.array([])
    yMat = np.array([])
    zMat = np.array([])
    tMat = np.array([], dtype=int)
    calSiteMat = []
    for i in range(dataSiteNum):
        if type(lineMat[i]) is int:
            continue
        peakX = lineMat[i].get_xdata()
        tMat = np.append(tMat, int(peakX[0]))
        siteName = drawSiteMat[i]
        calSiteMat.append(siteName)
        siteIndex = cAllSite.index(siteName)
        xMat = np.append(xMat, cAllX[siteIndex])
        yMat = np.append(yMat, cAllY[siteIndex])
        zMat = np.append(zMat, cAllZ[siteIndex])

    # #Main site
    for trySite in mainSiteMat:
        if trySite in calSiteMat:
            mainSite = trySite
            break
    mainSiteIndex = drawSiteMat.index(mainSite)
    mainPeakIndex = tMat[calSiteMat.index(mainSite)]
    mainPulse = dataMatList[mainSiteIndex][mainPeakIndex-upWidth:mainPeakIndex+downWidth]
    #Other sites
    for i in range(len(calSiteMat)):
        siteName = calSiteMat[i]
        if siteName==mainSite:
            continue
        siteIndex = drawSiteMat.index(siteName)
        oldPeak = int(tMat[i])
        maxCorr = 0
        for tryPeak in range(oldPeak-slideLen, oldPeak+slideLen):
            tryPulse = dataMatList[siteIndex][tryPeak-upWidth:tryPeak+downWidth]
            corrMat = np.corrcoef(mainPulse, tryPulse)
            corrVal = corrMat[0, 1]
            if corrVal>maxCorr:
                maxCorr = corrVal
                tMat[i] = tryPeak
    
    if loc2D==1:
        [resMat, resiMat, aveResi] = locFun2D(xMat, yMat, zMat, tMat)
        heightRes = 1000
    else:
        [resMat, resiMat, aveResi] = locFun(xMat, yMat, zMat, tMat)
        heightRes = resMat[3]
    
    if year==2021:
        peakCurrent = calPeak2021(resMat[1], resMat[0])
    else:
        peakCurrent = calPeak(resMat[1], resMat[0])
        
    print(resMat[0:2], aveResi)
    [xVal, yVal] = convertXY(resMat[0], resMat[1])
    resFile = resDir+'manualLoc_'+strTime+'.dat'
    if os.path.exists(resFile)==False:
        fidRes = open(resFile, 'at')
        fidRes.write(strNatTime)
        for siteName in drawSiteMat:
            fidRes.write(' '+siteName+',')
        fidRes.write('\n')
        fidRes.close()
    fidRes = open(resFile, 'at')
    fidRes.write('%10.4f,%7d,%7d,%9.5f,%10.5f,%5d,%4d,%2d,%3d,%3d' % (resMat[2]/1e3/lengthFactor, round(xVal), round(yVal), resMat[1]/np.pi*180, resMat[0]/np.pi*180, heightRes, peakCurrent, len(calSiteMat), aveResi, max(resiMat)))
    for i in range(dataSiteNum):
        siteName = drawSiteMat[i]
        if siteName in calSiteMat:
            siteIndex = calSiteMat.index(siteName)
            calPeakIndex = tMat[siteIndex]
            fidRes.write(',%9d' % calPeakIndex)
        else:
            fidRes.write(',%9d' % -1)
    fidRes.write('\n')
    fidRes.close()
    
def dis(x0, y0, xi, yi):
    return R*np.arccos(np.sin(x0)*np.sin(xi)+np.cos(x0)*np.cos(xi)*np.cos(y0-yi))

        
sampRate = oriDataLen
def key_press_fcn(event):
    strKey = event.key
    curX = event.xdata
    curY = event.ydata
    if strKey == '1':
        drawPro(curX, int(round(sampRate*0.05)))
    elif strKey == '2':
        drawPro(curX, int(round(sampRate*0.005)))
    elif strKey == '3':
        drawPro(curX, int(round(sampRate*0.0005)))
        plotLine()
    elif strKey == '4':
        drawPro(oriDataLen/2, oriDataLen/2)
    elif strKey == '5':
        drawPro(curX, int(round(sampRate*0.0001)))
    elif strKey == 'e':
        yIncreAll()
    elif strKey == 'w':
        yIncre(event.inaxes, curY)
    elif strKey == 'p':
        chanPola()
    elif strKey == 'x':
        delLine(event.inaxes)
    elif strKey == 't':
        print(curX, curX/1000.0)
    elif strKey == 'c':
        calLoc()
            

def click_fcn(event):
    try:
        curX = int(event.xdata)
        axIndex = axMat.index(event.inaxes)
    except:
        return
    if type(lineMat[axIndex]) is int:
        return
    dataSeg = dataMatList[axIndex][curX-peakWidth:curX+peakWidth]
    if linePola>0:
        newPeakX = dataSeg.argmax()+curX-peakWidth
    else:
        newPeakX = dataSeg.argmin()+curX-peakWidth
    lineMat[axIndex].set_xdata([newPeakX, newPeakX])
    f.canvas.draw()

f.canvas.mpl_connect('key_press_event', key_press_fcn)
f.canvas.mpl_connect('button_press_event', click_fcn)

readData()
drawPro(oriDataLen/2, oriDataLen/2)
plt.show()
